from ipaddress import IPv4Network, IPv6Network

from .roa_trie import ROATrie


class IPv4ROATrie(ROATrie[IPv4Network]):  # type: ignore
    """Trie of IPv4 CIDRs for ROAs"""

    PrefixCls = IPv4Network


class IPv6ROATrie(ROATrie[IPv6Network]):  # type: ignore
    """Trie of IPv6 CIDRs for ROAs"""

    PrefixCls = IPv6Network



from lib_cidr_trie import CIDRTrie
from lib_cidr_trie.cidr_trie import PrefixType

from .roa import ROA
from .enums import ROARouted, ROAValidity


class ROATrie(CIDRTrie[PrefixType]):
    """Trie of CIDRs for ROAs"""

    # Just changing the defaults here to ROA instead of CIDRNode
    def __init__(self, *args, NodeCls: type[ROA] = ROA, **kwargs) -> None:
        # mypy doesn't understand that I can reset the default arg in this way
        super(ROATrie, self).__init__(*args, NodeCls=NodeCls, **kwargs)  # type: ignore

    def get_validity(
        self, prefix: PrefixType, origin: int
    ) -> tuple[ROAValidity, ROARouted]:
        """Gets the validity of a prefix-origin pair"""

        roa = self.get_most_specific_trie_supernet(prefix)
        assert roa is None or isinstance(roa, ROA), "mypy type check"
        if roa:
            return roa.get_validity(prefix, origin)
        else:
            return ROAValidity.UNKNOWN, ROARouted.UNKNOWN



from ipaddress import IPv4Network, IPv6Network
from typing import Optional

from lib_cidr_trie import CIDRTrie

from .roa import ROA
from .roa_trie import ROATrie
from .roa_tries import IPv4ROATrie, IPv6ROATrie
from .enums import ROARouted, ROAValidity


class ROAChecker:
    """Gets validity of prefix origin pairs against ROAs"""

    def __init__(self):
        """Initializes both ROA tries"""

        self.ipv4_trie = IPv4ROATrie()
        self.ipv6_trie = IPv6ROATrie()

    def insert(
        self, prefix: IPv4Network | IPv6Network, origin: int, max_length: Optional[int]
    ) -> None:
        """Inserts a prefix into the tries"""

        trie = self.ipv4_trie if prefix.version == 4 else self.ipv6_trie
        # mypy struggling with this
        return trie.insert(prefix, origin, max_length)  # type: ignore

    def get_roa(self, prefix: IPv4Network | IPv6Network, *args) -> Optional[ROA]:
        """Gets the ROA covering prefix-origin pair"""

        trie = self.ipv4_trie if prefix.version == 4 else self.ipv6_trie
        assert isinstance(trie, CIDRTrie)
        roa = trie.get_most_specific_trie_supernet(prefix)
        assert roa is None or isinstance(roa, ROA), "for mypy"
        return roa

    def get_validity(
        self, prefix: IPv4Network | IPv6Network, origin: int
    ) -> tuple[ROAValidity, ROARouted]:
        """Gets the validity of a prefix origin pair"""

        trie = self.ipv4_trie if prefix.version == 4 else self.ipv6_trie
        assert isinstance(trie, ROATrie), "for mypy"
        return trie.get_validity(prefix, origin)



from enum import Enum


class ROAValidity(Enum):
    VALID = 0
    UNKNOWN = 1
    # Note that we cannot differentiate between invalid by length
    # or invalid by origin or invalid by both
    # That is because for the same prefix you can have multiple
    # max lengths or multiple origins
    # And you select the most valid roa. So is invalid by length
    # more valid than invalid by origin? No. So we just say invalid

    # Nixing the comment above with the following methodology:
    # NOTE: There can be multiple ROAs for the same prefix
    # So if we say a ROA is invalid by length and origin
    # it could potentially be invalid by length for one ROA
    # and invalid by origin for another prefix
    # If we say non routed, it's violating at least one non routed ROA
    INVALID_LENGTH = 2
    INVALID_ORIGIN = 3
    INVALID_LENGTH_AND_ORIGIN = 4

    @staticmethod
    def is_valid(roa_validity: "ROAValidity") -> bool:
        return roa_validity == ROAValidity.VALID

    @staticmethod
    def is_unknown(roa_validity: "ROAValidity") -> bool:
        return roa_validity == ROAValidity.UNKNOWN

    @staticmethod
    def is_invalid(roa_validity: "ROAValidity") -> bool:
        return roa_validity in (
            ROAValidity.INVALID_LENGTH,
            ROAValidity.INVALID_ORIGIN,
            ROAValidity.INVALID_LENGTH_AND_ORIGIN,
        )


class ROARouted(Enum):
    ROUTED = 0
    UNKNOWN = 1
    # A ROA is Non Routed if it is for an origin of ASN 0
    # This means that the prefix for this ROA should never be announced
    NON_ROUTED = 2



from ipaddress import IPv4Network, IPv6Network
from typing import Optional

from lib_cidr_trie import CIDRNode

from .enums import ROARouted, ROAValidity


class ROA(CIDRNode):
    def __init__(self, *args, **kwargs):
        """Initializes the ROA node"""

        super(ROA, self).__init__(*args, **kwargs)
        # Origin max length pairs
        self.origin_max_lengths: set[tuple[int, int]] = set()

    # Mypy doesn't understand *args in super class
    def add_data(  # type: ignore
        self, prefix: IPv4Network | IPv6Network, origin: int, max_length: Optional[int]
    ):
        """Adds data to the node"""

        if max_length is None:
            max_length = prefix.prefixlen
        self.prefix = prefix
        self.origin_max_lengths.add((origin, max_length))

    def get_validity(
        self, prefix: IPv4Network | IPv6Network, origin: int
    ) -> tuple[ROAValidity, ROARouted]:
        """Gets the ROA validity of a prefix origin pair"""

        assert isinstance(prefix, type(self.prefix))
        # Mypy isn't getting that these types are the same
        if not prefix.subnet_of(self.prefix):  # type: ignore
            return ROAValidity.UNKNOWN, ROARouted.UNKNOWN
        else:
            valid_length = True
            valid_origin = True
            routed = ROARouted.ROUTED
            # NOTE: There can be multiple ROAs for the same prefix
            # So if we say a ROA is invalid by length and origin
            # it could potentially be invalid by length for one ROA
            # and invalid by origin for another prefix
            # If we say non routed, it's violating at least one non routed ROA
            for self_origin, max_length in self.origin_max_lengths:
                if prefix.prefixlen > max_length:
                    valid_length = False
                    if self_origin == 0:
                        routed = ROARouted.NON_ROUTED
                if origin != self_origin:
                    valid_origin = False
                    if self_origin == 0:
                        routed = ROARouted.NON_ROUTED
            if valid_length and valid_origin:
                return ROAValidity.VALID, routed
            elif valid_length and not valid_origin:
                return ROAValidity.INVALID_ORIGIN, routed
            elif not valid_length and valid_origin:
                return ROAValidity.INVALID_LENGTH, routed
            else:
                return ROAValidity.INVALID_LENGTH_AND_ORIGIN, routed



from ipaddress import ip_network

from roa_checker import ROAChecker, ROARouted, ROAValidity


def test_tree():
    # TODO: Break up into unit tests
    trie = ROAChecker()
    cidrs = [ip_network(x) for x in ["1.2.0.0/16", "1.2.3.0/24", "1.2.3.4"]]
    routed_origin = 1
    for cidr in cidrs:
        trie.insert(cidr, routed_origin, cidr.prefixlen)
    for cidr in cidrs:
        roa = trie.get_roa(cidr, routed_origin)
        assert roa and roa.prefix == cidr
        validity, routed = trie.get_validity(cidr, routed_origin)
        assert (validity, routed) == (
            ROAValidity.VALID,
            ROARouted.ROUTED,
        )
        assert ROAValidity.is_unknown(validity) is False
        assert ROAValidity.is_invalid(validity) is False
        assert ROAValidity.is_valid(validity) is True

    non_routed_cidrs = [ip_network(x) for x in ["2.2.0.0/16", "2.2.3.0/24", "2.2.3.4"]]
    non_routed_origin = 0
    for cidr in non_routed_cidrs:
        trie.insert(cidr, non_routed_origin, cidr.prefixlen)
    for cidr in non_routed_cidrs:
        roa = trie.get_roa(cidr, non_routed_origin)
        assert roa and roa.prefix == cidr
        assert trie.get_validity(cidr, routed_origin) == (
            ROAValidity.INVALID_ORIGIN,
            ROARouted.NON_ROUTED,
        )

    validity, routed = trie.get_validity(ip_network("1.0.0.0/8"), routed_origin)
    assert validity == ROAValidity.UNKNOWN
    assert routed == ROARouted.UNKNOWN
    validity, routed = trie.get_validity(ip_network("255.255.255.255"), routed_origin)
    assert validity == ROAValidity.UNKNOWN
    assert routed == ROARouted.UNKNOWN
    assert ROAValidity.is_unknown(validity) is True
    assert ROAValidity.is_invalid(validity) is False
    assert ROAValidity.is_valid(validity) is False
    validity, routed = trie.get_validity(ip_network("1.2.4.0/24"), routed_origin)
    assert validity == ROAValidity.INVALID_LENGTH
    assert routed == ROARouted.ROUTED
    assert ROAValidity.is_unknown(validity) is False
    assert ROAValidity.is_invalid(validity) is True
    assert ROAValidity.is_valid(validity) is False
    validity, routed = trie.get_validity(ip_network("1.2.3.0/24"), routed_origin + 1)
    assert validity == ROAValidity.INVALID_ORIGIN
    assert routed == ROARouted.ROUTED
    assert ROAValidity.is_unknown(validity) is False
    assert ROAValidity.is_invalid(validity) is True
    assert ROAValidity.is_valid(validity) is False
    validity, routed = trie.get_validity(ip_network("1.2.4.0/24"), routed_origin + 1)
    assert validity == ROAValidity.INVALID_LENGTH_AND_ORIGIN
    assert routed == ROARouted.ROUTED
    assert ROAValidity.is_unknown(validity) is False
    assert ROAValidity.is_invalid(validity) is True
    assert ROAValidity.is_valid(validity) is False
    validity, routed = trie.get_validity(ip_network("1.2.0.255"), routed_origin)
    assert validity == ROAValidity.INVALID_LENGTH
    assert routed == ROARouted.ROUTED
    validity, routed = trie.get_validity(ip_network("1.3.0.0/16"), routed_origin)
    assert validity == ROAValidity.UNKNOWN
    assert routed == ROARouted.UNKNOWN
    validity, routed = trie.get_validity(ip_network("1.2.0.255"), routed_origin)
    assert validity == ROAValidity.INVALID_LENGTH
    assert routed == ROARouted.ROUTED
